--[[ HatMenu 1.0 ]]--

HatMenu = {}

-- localized strings required to support engineering bags
HatMenu.BAG = "Bag" -- 7th return of GetItemInfo on a normal bag
HatMenu.ENGINEERING_BAG = "Engineering Bag" -- 7th return of GetItemInfo on an engineering bag
HatMenu.TRADE_GOODS = "Trade Goods" -- 6th return of GetItemInfo on most engineered hats
HatMenu.DEVICES = "Devices" -- 7th return of GetItemInfo on most engineered hats
HatMenu.REQUIRES_ENGINEERING = "Requires Engineering" -- from tooltip when GetItemInfo ambiguous

function HatMenu.LoadDefaults()

	HatMenuOptions = HatMenuOptions or {
		IconPos = -100,			-- angle of initial minimap icon position
		ShowIcon = "ON",		-- whether to show the minimap button
		SquareMinimap = "OFF",		-- whether the minimap is square instead of circular
		CooldownCount = "OFF",		-- whether to display numerical cooldown counters
		LargeCooldown = "ON",		-- whether cooldown numbers are large or small
		TooltipFollow = "OFF",		-- whether tooltips follow the mouse
		KeepOpen = "OFF",		-- whether menu hides after use
		KeepDocked = "ON",		-- whether to keep menu docked at all times
		Notify = "OFF",			-- whether a message appears when a hat is ready
		DisableToggle="OFF",		-- whether minimap button toggles hat frame
		NotifyUsedOnly="OFF",		-- whether notify happens only on hats used
		NotifyChatAlso="OFF",		-- whether to send notify to chat also
		Locked = "OFF",			-- whether windows can be moved/scaled/rotated
		ShowTooltips = "ON",		-- whether to display tooltips at all
		NotifyThirty = "OFF",		-- whether to notify cooldowns at 30 seconds instead of 0
		MenuOnShift = "OFF",		-- whether menu requires Shift to display
		TinyTooltips = "OFF",		-- whether tooltips display only name and cooldown
		SetColumns = "OFF",		-- whether number of columns in menu is chosen automatically
		Columns = 4,			-- if SetColumns "ON", number of columns before menu wraps
		ShowHotKeys = "OFF",		-- whether hotkeys show on hats
		RedRange = "OFF",		-- whether to monitor and red out out of range hats
		MenuOnRight = "OFF",		-- whether to open menu with right-click
	}

	HatMenuPerOptions = HatMenuPerOptions or {
		MainDock = "BOTTOMRIGHT",	-- corner of main window docked to
		MenuDock = "BOTTOMLEFT",	-- corner menu window is docked from
		MenuOrient = "VERTICAL",	-- direction of menu window
		XPos = 480,			-- left edge of main window
		YPos = 400,			-- top edge of main window
		MainScale = 1,			-- scaling of main window
		MenuScale = 1,			-- scaling of menu window
		Visible="ON",			-- whether to display the hats
		FirstUse = true,		-- whether this is the first time this user has used the mod
		ItemsUsed = {},			-- table of hats used and their cooldown status
		Alpha = 1,			-- alpha of both windows
		Hidden = {},			-- table of hats hidden
	}
end

--[[ Misc Variables ]]--

HatMenu_Version = "1.0"
BINDING_HEADER_HATMENU = "HatMenu"

HatMenu.MaxHats = 30 -- add more to HatMenu_MenuFrame if this changes
HatMenu.BaggedHats = {} -- indexed by number, 1-30 of hats in the menu
HatMenu.NumberOfHats = 0 -- number of hats in the menu
HatMenu.CombatQueue = {} -- [0] or [1] = name of hat queued for slot 0 or 1
HatMenu.Corners = {"TOPLEFT", "TOPRIGHT", "BOTTOMLEFT", "BOTTOMRIGHT"}

HatMenu.Buffs = {} -- buffs currently on player, indexed by buff name

--[[ Local functions ]]--

-- dock-dependant offset and directions: MainDock..MenuDock
-- x/yoff   = offset MenuFrame is positioned to MainFrame
-- x/ydir   = direction hats are added to menu
-- x/ystart = starting offset when building a menu, relativePoint MenuDock
HatMenu.DockStats = {	["TOPRIGHTTOPLEFT"] =		{ xoff=-4, yoff=0,  xdir=1,  ydir=-1, xstart=8,   ystart=-8 },
			["BOTTOMRIGHTBOTTOMLEFT"] =	{ xoff=-4, yoff=0,  xdir=1,  ydir=1,  xstart=8,   ystart=44 },
			["TOPLEFTTOPRIGHT"] =		{ xoff=4,  yoff=0,  xdir=-1, ydir=-1, xstart=-44, ystart=-8 },
			["BOTTOMLEFTBOTTOMRIGHT"] =	{ xoff=4,  yoff=0,  xdir=-1, ydir=1,  xstart=-44, ystart=44 },
			["TOPRIGHTBOTTOMRIGHT"] =	{ xoff=0,  yoff=-4, xdir=-1, ydir=1,  xstart=-44, ystart=44 },
			["BOTTOMRIGHTTOPRIGHT"] =	{ xoff=0,  yoff=4,  xdir=-1, ydir=-1, xstart=-44, ystart=-8 },
			["TOPLEFTBOTTOMLEFT"] =		{ xoff=0,  yoff=-4, xdir=1,  ydir=1,  xstart=8,   ystart=44 },
			["BOTTOMLEFTTOPLEFT"] =		{ xoff=0,  yoff=4,  xdir=1,  ydir=-1, xstart=8,   ystart=-8 } }

-- returns offset and direction depending on current docking. ie: HatMenu.DockInfo("xoff")
function HatMenu.DockInfo(arg1)
	local anchor = HatMenuPerOptions.MainDock..HatMenuPerOptions.MenuDock
	if HatMenu.DockStats[anchor] and arg1 and HatMenu.DockStats[anchor][arg1] then
		return HatMenu.DockStats[anchor][arg1]
	else
		return 0
	end
end

-- hide the docking markers
function HatMenu.ClearDocking()
	for i=1,4 do
		getglobal("HatMenu_MainDock_"..HatMenu.Corners[i]):Hide()
		getglobal("HatMenu_MenuDock_"..HatMenu.Corners[i]):Hide()
	end
end

-- returns true if the two values are close to each other
function HatMenu.Near(arg1,arg2)
	return (math.max(arg1,arg2)-math.min(arg1,arg2))<15
end

-- moves the MenuFrame to the dock position against MainFrame
function HatMenu.DockWindows()
	HatMenu.ClearDocking()
	if HatMenuOptions.KeepDocked=="ON" then
		HatMenu_MenuFrame:ClearAllPoints()
		if HatMenuOptions.Locked=="OFF" then
			HatMenu_MenuFrame:SetPoint(HatMenuPerOptions.MenuDock,"HatMenu_MainFrame",HatMenuPerOptions.MainDock,HatMenu.DockInfo("xoff"),HatMenu.DockInfo("yoff"))
		else
			HatMenu_MenuFrame:SetPoint(HatMenuPerOptions.MenuDock,"HatMenu_MainFrame",HatMenuPerOptions.MainDock,HatMenu.DockInfo("xoff")*3,HatMenu.DockInfo("yoff")*3)
		end
	end
	if HatMenu_MenuFrame:IsVisible() then
		HatMenu.BuildMenu()
	end
end

-- scan inventory and build MenuFrame
function HatMenu.BuildMenu()

	if not IsShiftKeyDown() and HatMenuOptions.MenuOnShift=="ON" then
		return
	end

	local idx,i,j,k,texture = 1
	local itemLink,itemID,itemName,equipSlot,itemTexture

	-- go through bags and gather hats into .BaggedHats
	for i=0,4 do
		for j=1,GetContainerNumSlots(i) do
			itemLink = GetContainerItemLink(i,j)
			
			if itemLink then
				_,_,itemID,itemName = string.find(GetContainerItemLink(i,j) or "","item:(%d+).+%[(.+)%]")
				_,_,_,_,_,_,_,_,equipSlot,itemTexture = GetItemInfo(itemID or "")
				if equipSlot=="INVTYPE_HEAD" and (IsAltKeyDown() or not HatMenuPerOptions.Hidden[itemID]) then
					if not HatMenu.BaggedHats[idx] then
						HatMenu.BaggedHats[idx] = {}
					end
					HatMenu.BaggedHats[idx].id = itemID
					HatMenu.BaggedHats[idx].bag = i
					HatMenu.BaggedHats[idx].slot = j
					HatMenu.BaggedHats[idx].name = itemName
					HatMenu.BaggedHats[idx].texture = itemTexture
					idx = idx + 1
				end
			end
		end
	end
	HatMenu.NumberOfHats = math.min(idx-1,HatMenu.MaxHats)

	if HatMenu.NumberOfHats<1 then
		-- user has no bagged hats :(
		HatMenu_MenuFrame:Hide()
	else
		-- display hats outward from docking point
		local col,row,xpos,ypos = 0,0,HatMenu.DockInfo("xstart"),HatMenu.DockInfo("ystart")
		local max_cols = 1

		if HatMenu.NumberOfHats>24 then
			max_cols = 5
		elseif HatMenu.NumberOfHats>18 then
			max_cols = 4
		elseif HatMenu.NumberOfHats>12 then
			max_cols = 3
		elseif HatMenu.NumberOfHats>4 then
			max_cols = 2
		end
		if HatMenuOptions.SetColumns=="ON" and HatMenuOptions.Columns then
			max_cols = HatMenuOptions.Columns
		end

		local item,icon
		for i=1,HatMenu.NumberOfHats do
			item = getglobal("HatMenu_Menu"..i)
			icon = getglobal("HatMenu_Menu"..i.."Icon")
			icon:SetTexture(HatMenu.BaggedHats[i].texture)
			if HatMenuPerOptions.Hidden[HatMenu.BaggedHats[i].id] then
				icon:SetDesaturated(1)
			else
				icon:SetDesaturated(0)
			end
			item:SetPoint("TOPLEFT","HatMenu_MenuFrame",HatMenuPerOptions.MenuDock,xpos,ypos)

			if HatMenuPerOptions.MenuOrient=="VERTICAL" then
				xpos = xpos + HatMenu.DockInfo("xdir")*40
				col = col + 1
				if col==max_cols then
					xpos = HatMenu.DockInfo("xstart")
					col = 0
					ypos = ypos + HatMenu.DockInfo("ydir")*40
					row = row + 1
				end
				item:Show()
			else
				ypos = ypos + HatMenu.DockInfo("ydir")*40
				col = col + 1
				if col==max_cols then
					ypos = HatMenu.DockInfo("ystart")
					col = 0
					xpos = xpos + HatMenu.DockInfo("xdir")*40
					row = row + 1
				end
				item:Show()
			end
		end
		for i=(HatMenu.NumberOfHats+1),HatMenu.MaxHats do
			getglobal("HatMenu_Menu"..i):Hide()
		end
		if col==0 then
			row = row-1
		end

		if HatMenuPerOptions.MenuOrient=="VERTICAL" then
			HatMenu_MenuFrame:SetWidth(12+(max_cols*40))
			HatMenu_MenuFrame:SetHeight(12+((row+1)*40))
		else
			HatMenu_MenuFrame:SetWidth(12+((row+1)*40))
			HatMenu_MenuFrame:SetHeight(12+(max_cols*40))
		end
		HatMenu.UpdateMenuCooldowns()
		HatMenu_MenuFrame:Show()
		HatMenu.StartTimer("MenuMouseover")
	end

end

function HatMenu.Initialize()

	local options = HatMenuOptions
	options.KeepDocked = options.KeepDocked or "ON" -- new option for 2.1
	options.Notify = options.Notify or "OFF" -- 2.1
	options.DisableToggle = options.DisableToggle or "OFF" -- new option for 2.2
	options.NotifyUsedOnly = options.NotifyUsedOnly or "OFF" -- 2.2
	options.NotifyChatAlso = options.NotifyChatAlso or "OFF" -- 2.2
	options.ShowTooltips = options.ShowTooltips or "ON" -- 2.3
	options.NotifyThirty = options.NotifyThirty or "OFF" -- 2.5
	options.SquareMinimap = options.SquareMinimap or "OFF" -- 2.6
	options.MenuOnShift = options.MenuOnShift or "OFF" -- 2.6
	options.TinyTooltips = options.TinyTooltips or "OFF" -- 3.0
	options.SetColumns = options.SetColumns or "OFF" -- 3.0
	options.Columns = options.Columns or 4 -- 3.0
	options.LargeCooldown = options.LargeCooldown or "OFF" -- 3.0
	options.ShowHotKeys = options.ShowHotKeys or "OFF" -- 3.0
	HatMenuPerOptions.ItemsUsed = HatMenuPerOptions.ItemsUsed or {} -- 3.0
	options.HideOnLoad = options.HideOnLoad or "OFF" -- 3.4
	options.RedRange = options.RedRange or "OFF" -- 3.54
	HatMenuPerOptions.Alpha = HatMenuPerOptions.Alpha or 1 -- 3.5
	HatMenuPerOptions.Hidden = HatMenuPerOptions.Hidden or {}
	options.MenuOnRight = options.MenuOnRight or "OFF" -- 3.61

	if HatMenuPerOptions.XPos and HatMenuPerOptions.YPos then
		HatMenu_MainFrame:SetPoint("TOPLEFT","UIParent","BOTTOMLEFT",HatMenuPerOptions.XPos,HatMenuPerOptions.YPos)
	end
	if HatMenuPerOptions.MainScale then
		HatMenu_MainFrame:SetScale(HatMenuPerOptions.MainScale)
	end
	if HatMenuPerOptions.MenuScale then
		HatMenu_MenuFrame:SetScale(HatMenuPerOptions.MenuScale)
	end
	HatMenu.ReflectAlpha()

	HatMenu_Hat:SetAttribute("type","item")
	HatMenu_Hat:SetAttribute("slot",1)
	HatMenu_Hat:SetAttribute("shift-slot*",ATTRIBUTE_NOOP)
	HatMenu.ReflectMenuOnRight()

	HatMenu.InitTimers()
	HatMenu.CreateTimer("UpdateWornHats",HatMenu.UpdateWornHats,.75)
	HatMenu.CreateTimer("DockingMenu",HatMenu.DockingMenu,.2,1)
	HatMenu.CreateTimer("MenuMouseover",HatMenu.MenuMouseover,.25,1)
	HatMenu.CreateTimer("Scaling",HatMenu.Scaling,.1,1)
	HatMenu.CreateTimer("TooltipUpdate",HatMenu.TooltipUpdate,1,1)
	HatMenu.CreateTimer("CooldownUpdate",HatMenu.CooldownUpdate,1,1)
	HatMenu.CreateTimer("RedRange",HatMenu.RedRangeUpdate,.33,1)

	hooksecurefunc("UseInventoryItem", HatMenu.newUseInventoryItem)
	hooksecurefunc("UseAction", HatMenu.newUseAction)

	HatMenu.InitOptions()

	HatMenu.UpdateWornHats()
	HatMenu.DockWindows()
	HatMenu.StartTimer("CooldownUpdate")
	HatMenu.ReflectRedRange()

	if HatMenuPerOptions.Visible=="ON" and (GetInventoryItemLink("player",1)) then
		HatMenu_MainFrame:Show()
	end

	-- fix for OmniCC by N00bZXI
	HatMenu_MainFrame:SetFrameLevel(1)
	HatMenu_MenuFrame:SetFrameLevel(1)
	HatMenu_Hat:SetFrameLevel(2)
	for i=1,30 do
		getglobal("HatMenu_Menu"..i):SetFrameLevel(2)
	end

end

function HatMenu.ReflectMenuOnRight()
	HatMenu_Hat:SetAttribute("slot2",HatMenuOptions.MenuOnRight=="ON" and ATTRIBUTE_NOOP or nil)
end

-- returns true if the player is really dead or ghost, not merely FD
function HatMenu.IsPlayerReallyDead()
	local dead = UnitIsDeadOrGhost("player")
	if select(2,UnitClass("player"))=="HUNTER" then
		for i=1,24 do
			if select(3,UnitBuff("player",i))=="Interface\\Icons\\Ability_Rogue_FeignDeath" then
				return nil
			end
		end
	end
	return dead
end

function HatMenu.ItemInfo()
	local link,id,name,equipLoc,texture = GetInventoryItemLink("player",1)
	if link then
		local _,_,id = string.find(link,"item:(%d+)")
		name,_,_,_,_,_,_,_,equipLoc,texture = GetItemInfo(id)
	else
		_,texture = GetInventorySlotInfo("HatSlot")
	end
	return texture,name,equipLoc
end

function HatMenu.FindItem(name,includeInventory)
	if includeInventory then
		if string.find(GetInventoryItemLink("player",1) or "",name,1,1) then
			return 1
		end
	end
	for i=0,4 do
		for j=1,GetContainerNumSlots(i) do
			if string.find(GetContainerItemLink(i,j) or "",name,1,1) then
				return nil,i,j
			end
		end
	end
end

--[[ Frame Scripts ]]--

function HatMenu.OnLoad()

	SlashCmdList["HatMenuCOMMAND"] = HatMenu.SlashHandler
	SLASH_HatMenuCOMMAND1 = "/hatmenu";
	SLASH_HatMenuCOMMAND2 = "/hat";
	
	this:RegisterEvent("PLAYER_LOGIN")
end

function HatMenu.OnEvent()

	if event=="UNIT_INVENTORY_CHANGED" and arg1=="player" then
		HatMenu.UpdateWornHats()
	elseif event=="ACTIONBAR_UPDATE_COOLDOWN" then
		HatMenu.UpdateWornCooldowns(1)
	elseif event=="PLAYER_AURAS_CHANGED" then
		HatMenu.UpdateBuffs()
	elseif (event=="PLAYER_REGEN_ENABLED" or event=="PLAYER_UNGHOST" or event=="PLAYER_ALIVE") and not HatMenu.IsPlayerReallyDead() then
		-- hats can now be swapped after combat/death
		if HatMenu.CombatQueue[0] then
			HatMenu.EquipHatByName(HatMenu.CombatQueue[0])
			HatMenu.CombatQueue[0] = nil
			HatMenu.UpdateCombatQueue()
		end
		HatMenu_OptMenuOnRight:Enable()
	elseif event=="UPDATE_BINDINGS" then
		HatMenu.KeyBindingsChanged()
	elseif event=="PLAYER_REGEN_DISABLED" then
		HatMenu_OptMenuOnRight:Disable()
	elseif event=="PLAYER_LOGIN" then
		HatMenu.LoadDefaults()
		HatMenu.Initialize()
		this:RegisterEvent("PLAYER_REGEN_ENABLED")
		this:RegisterEvent("PLAYER_REGEN_DISABLED")
		this:RegisterEvent("PLAYER_UNGHOST")
		this:RegisterEvent("PLAYER_ALIVE")
		this:RegisterEvent("UNIT_INVENTORY_CHANGED")
		this:RegisterEvent("UPDATE_BINDINGS")
		this:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
	end
end

function HatMenu.UpdateWornHats()
	local texture,name = HatMenu.ItemInfo()
	HatMenu_HatIcon:SetTexture(texture)
	HatMenu_HatIcon:SetDesaturated(0)
	HatMenu_Hat:SetChecked(0)
	HatMenu.UpdateWornCooldowns()
	if HatMenu_MenuFrame:IsVisible() then
		HatMenu.BuildMenu()
	end
end

function HatMenu.SlashHandler(msg)

	msg = string.lower(msg)

	if not msg or msg=="" then
		HatMenu.ToggleFrame(HatMenu_MainFrame)
	elseif string.find(msg,"^opt") or string.find(msg,"^config") then
		HatMenu.ToggleFrame(HatMenu_OptFrame)
	elseif msg=="lock" then
		HatMenuOptions.Locked="ON"
		HatMenu.DockWindows()
		HatMenu.ReflectLock()
	elseif msg=="unlock" then
		HatMenuOptions.Locked="OFF"
		HatMenu.DockWindows()
		HatMenu.ReflectLock()
	elseif msg=="reset" then
		HatMenu.ResetSettings()
	elseif string.find(msg,"alpha") then
		local _,_,alpha = string.find(msg,"alpha (.+)")
		alpha = tonumber(alpha)
		if alpha and alpha>0 and alpha<=1.0 then
			HatMenuPerOptions.Alpha = alpha
		else
			DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00HatMenu alpha:")
			DEFAULT_CHAT_FRAME:AddMessage("hat alpha (number) : set alpha from 0.1 to 1.0")
		end
		HatMenu.ReflectAlpha()
	elseif string.find(msg,"scale") then
		local _,_,menuscale = string.find(msg,"scale menu (.+)")
		if tonumber(menuscale) then
			HatMenu.FrameToScale = HatMenu_MenuFrame
			HatMenu.ScaleFrame(menuscale)
		end
		local _,_,mainscale = string.find(msg,"scale main (.+)")
		if tonumber(mainscale) then
			HatMenu.FrameToScale = HatMenu_MainFrame
			HatMenu.ScaleFrame(mainscale)
		end
		if not tonumber(menuscale) and not tonumber(mainscale) then
			DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00HatMenu scale:")
			DEFAULT_CHAT_FRAME:AddMessage("/hat scale main (number) : set exact main scale")
			DEFAULT_CHAT_FRAME:AddMessage("/hat scale menu (number) : set exact menu scale")
			DEFAULT_CHAT_FRAME:AddMessage("ie, /hat scale menu 0.85")
			DEFAULT_CHAT_FRAME:AddMessage("Note: You can drag the lower-right corner of either window to scale.  This slash command is for those who want to set an exact scale.")
		end
		HatMenu.FrameToScale = nil
		HatMenuPerOptions.MainScale = HatMenu_MainFrame:GetScale()
		HatMenuPerOptions.MenuScale = HatMenu_MenuFrame:GetScale()
	else
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFFF00HatMenu useage:")
		DEFAULT_CHAT_FRAME:AddMessage("/hat or /hatmenu : toggle the window")
		DEFAULT_CHAT_FRAME:AddMessage("/hat reset : reset all settings")
		DEFAULT_CHAT_FRAME:AddMessage("/hat opt : summon options window")
		DEFAULT_CHAT_FRAME:AddMessage("/hat lock|unlock : toggles window lock")
		DEFAULT_CHAT_FRAME:AddMessage("/hat scale main|menu (number) : sets an exact scale")
	end
end

function HatMenu.ResetSettings()
	StaticPopupDialogs["HATMENURESET"] = {
		text = "Are you sure you want to reset HatMenu to default state and reload the UI?",
		button1 = "Yes", button2 = "No", showAlert=1, timeout = 0, whileDead = 1,
		OnAccept = function() HatMenuOptions=nil HatMenuPerOptions=nil ReloadUI() end,
	}
	StaticPopup_Show("HATMENURESET")
end

--[[ Window Movement ]]--

function HatMenu.MainFrame_OnMouseUp()
	if arg1=="LeftButton" then
		this:StopMovingOrSizing()
		HatMenuPerOptions.XPos = HatMenu_MainFrame:GetLeft()
		HatMenuPerOptions.YPos = HatMenu_MainFrame:GetTop()
	end

end

function HatMenu.MainFrame_OnMouseDown(arg1)
	if arg1=="LeftButton" and HatMenuOptions.Locked=="OFF" then
		this:StartMoving()
	end
end

--[[ Timers ]]

function HatMenu.InitTimers()
	HatMenu.TimerPool = {}
	HatMenu.Timers = {}
end

function HatMenu.CreateTimer(name,func,delay,rep)
	HatMenu.TimerPool[name] = { func=func,delay=delay,rep=rep,elapsed=delay }
end

function HatMenu.IsTimerActive(name)
	for i,j in ipairs(HatMenu.Timers) do
		if j==name then
			return i
		end
	end
	return nil
end

function HatMenu.StartTimer(name,delay)
	HatMenu.TimerPool[name].elapsed = delay or HatMenu.TimerPool[name].delay
	if not HatMenu.IsTimerActive(name) then
		table.insert(HatMenu.Timers,name)
		HatMenu_TimersFrame:Show()
	end
end

function HatMenu.StopTimer(name)
	local idx = HatMenu.IsTimerActive(name)
	if idx then
		table.remove(HatMenu.Timers,idx)
		if #(HatMenu.Timers)<1 then
			HatMenu_TimersFrame:Hide()
		end
	end
end

function HatMenu.TimersFrame_OnUpdate()
	local timerPool
	for _,name in ipairs(HatMenu.Timers) do
		timerPool = HatMenu.TimerPool[name]
		timerPool.elapsed = timerPool.elapsed - arg1
		if timerPool.elapsed < 0 then
			timerPool.func()
			if timerPool.rep then
				timerPool.elapsed = timerPool.delay
			else
				HatMenu.StopTimer(name)
			end
		end
	end
end

function HatMenu.TimerDebug()
	local on = "|cFF00FF00On"
	local off = "|cFFFF0000Off"
	DEFAULT_CHAT_FRAME:AddMessage("|cFF44AAFFHatMenu_TimersFrame is "..(HatMenu_TimersFrame:IsVisible() and on or off))
	for i in pairs(HatMenu.TimerPool) do
		DEFAULT_CHAT_FRAME:AddMessage(i.." is "..(HatMenu.IsTimerActive(i) and on or off))
	end
end

--[[ OnClicks ]]

function HatMenu.MainHat_OnClick()
	this:SetChecked(0)
	if arg1=="RightButton" and HatMenuOptions.MenuOnRight=="ON" then
		if HatMenu_MenuFrame:IsVisible() then
			HatMenu_MenuFrame:Hide()
		else
			HatMenu.BuildMenu()
		end
	elseif IsShiftKeyDown() then
		if ChatFrameEditBox:IsVisible() then
			ChatFrameEditBox:Insert(GetInventoryItemLink("player",1))
		end
	else
		HatMenu.ReflectHatUse()
--		UseInventoryItem(this:GetID())
	end
end

function HatMenu.MenuHat_OnClick()
	this:SetChecked(0)
	local bag,slot = HatMenu.BaggedHats[this:GetID()].bag
	local slot = HatMenu.BaggedHats[this:GetID()].slot
	if IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
		ChatFrameEditBox:Insert(GetContainerItemLink(bag,slot))
	elseif IsAltKeyDown() then
		local _,_,itemID = string.find(GetContainerItemLink(bag,slot) or "","item:(%d+)")
		if HatMenuPerOptions.Hidden[itemID] then
			HatMenuPerOptions.Hidden[itemID] = nil
		else
			HatMenuPerOptions.Hidden[itemID] = 1
		end
		HatMenu.BuildMenu()
	else
		HatMenu.EquipHatByName(HatMenu.BaggedHats[this:GetID()].name)
		if not IsShiftKeyDown() and HatMenuOptions.KeepOpen=="OFF" then
			HatMenu_MenuFrame:Hide()
		end
	end
end

--[[ Docking ]]

function HatMenu.MenuFrame_OnMouseDown()
	if arg1=="LeftButton" and HatMenuOptions.Locked=="OFF" then
		HatMenu_MenuFrame:StartMoving()

		if HatMenuOptions.KeepDocked=="ON" then
			HatMenu.StartTimer("DockingMenu")
		end
	end
end

function HatMenu.MenuFrame_OnMouseUp()
	if arg1=="LeftButton" then
		HatMenu.StopTimer("DockingMenu")
		HatMenu_MenuFrame:StopMovingOrSizing()
		if HatMenuOptions.KeepDocked=="ON" then
			HatMenu.DockWindows()
		end
	elseif HatMenuOptions.Locked=="OFF" then
		if HatMenuPerOptions.MenuOrient=="VERTICAL" then
			HatMenuPerOptions.MenuOrient="HORIZONTAL"
		else
			HatMenuPerOptions.MenuOrient="VERTICAL"
		end
		HatMenu.BuildMenu()
	end
end

function HatMenu.DockingMenu()

	local main = HatMenu_MainFrame
	local menu = HatMenu_MenuFrame
	local mainscale = HatMenu_MainFrame:GetScale()
	local menuscale = HatMenu_MenuFrame:GetScale()
	local near = HatMenu.Near

	if near(main:GetRight()*mainscale,menu:GetLeft()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetTop()*menuscale) then
			HatMenuPerOptions.MainDock = "TOPRIGHT"
			HatMenuPerOptions.MenuDock = "TOPLEFT"
		elseif near(main:GetBottom()*mainscale,menu:GetBottom()*menuscale) then
			HatMenuPerOptions.MainDock = "BOTTOMRIGHT"
			HatMenuPerOptions.MenuDock = "BOTTOMLEFT"
		end
	elseif near(main:GetLeft()*mainscale,menu:GetRight()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetTop()*menuscale) then
			HatMenuPerOptions.MainDock = "TOPLEFT"
			HatMenuPerOptions.MenuDock = "TOPRIGHT"
		elseif near(main:GetBottom()*mainscale,menu:GetBottom()*menuscale) then
			HatMenuPerOptions.MainDock = "BOTTOMLEFT"
			HatMenuPerOptions.MenuDock = "BOTTOMRIGHT"
		end
	elseif near(main:GetRight()*mainscale,menu:GetRight()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetBottom()*menuscale) then
			HatMenuPerOptions.MainDock = "TOPRIGHT"
			HatMenuPerOptions.MenuDock = "BOTTOMRIGHT"
		elseif near(main:GetBottom()*mainscale,menu:GetTop()*menuscale) then
			HatMenuPerOptions.MainDock = "BOTTOMRIGHT"
			HatMenuPerOptions.MenuDock = "TOPRIGHT"
		end
	elseif near(main:GetLeft()*mainscale,menu:GetLeft()*menuscale) then
		if near(main:GetTop()*mainscale,menu:GetBottom()*menuscale) then
			HatMenuPerOptions.MainDock = "TOPLEFT"
			HatMenuPerOptions.MenuDock = "BOTTOMLEFT"
		elseif near(main:GetBottom()*mainscale,menu:GetTop()*menuscale) then
			HatMenuPerOptions.MainDock = "BOTTOMLEFT"
			HatMenuPerOptions.MenuDock = "TOPLEFT"
		end
	end
	HatMenu.ClearDocking()
	getglobal("HatMenu_MainDock_"..HatMenuPerOptions.MainDock):Show()
	getglobal("HatMenu_MenuDock_"..HatMenuPerOptions.MenuDock):Show()
end

function HatMenu.MenuMouseover()
	if (not MouseIsOver(HatMenu_MainFrame)) and (not MouseIsOver(HatMenu_MenuFrame)) and not IsShiftKeyDown() and (HatMenuOptions.KeepOpen=="OFF") and not HatMenu.IsTimerActive("Scaling") then
		HatMenu.StopTimer("MenuMouseover")
		HatMenu_MenuFrame:Hide()
	end
end

--[[ Scaling ]]

function HatMenu.StartScaling()
	if arg1=="LeftButton" and HatMenuOptions.Locked=="OFF" then
		this:LockHighlight()
		HatMenu.FrameToScale = this:GetParent()
		HatMenu.ScalingWidth = this:GetParent():GetWidth()
		HatMenu.StartTimer("Scaling")
	end
end

function HatMenu.StopScaling()
	if arg1=="LeftButton" then
		HatMenu.StopTimer("Scaling")
		HatMenu.FrameToScale = nil
		this:UnlockHighlight()
		if this:GetParent():GetName() == "HatMenu_MainFrame" then
			HatMenuPerOptions.MainScale = HatMenu_MainFrame:GetScale()
		else
			HatMenuPerOptions.MenuScale = HatMenu_MenuFrame:GetScale()
		end
	end
end

function HatMenu.ScaleFrame(scale)
	local frame = HatMenu.FrameToScale
	local oldscale = frame:GetScale() or 1
	local framex = (frame:GetLeft() or HatMenuPerOptions.XPos)* oldscale
	local framey = (frame:GetTop() or HatMenuPerOptions.YPos)* oldscale

	frame:SetScale(scale)
	if frame:GetName() == "HatMenu_MainFrame" then
		HatMenu_MainFrame:SetPoint("TOPLEFT","UIParent","BOTTOMLEFT",framex/scale,framey/scale)
		HatMenuPerOptions.XPos = HatMenu_MainFrame:GetLeft()
		HatMenuPerOptions.YPos = HatMenu_MainFrame:GetTop()
	elseif HatMenuOptions.KeepDocked=="OFF" then
		HatMenu_MenuFrame:ClearAllPoints()
		HatMenu_MenuFrame:SetPoint("TOPLEFT","UIParent","BOTTOMLEFT",framex/scale,framey/scale)
	end
end

function HatMenu.Scaling()
	local frame = HatMenu.FrameToScale
	local oldscale = frame:GetEffectiveScale()
	local framex, framey, cursorx, cursory = frame:GetLeft()*oldscale, frame:GetTop()*oldscale, GetCursorPosition()
	if (cursorx-framex)>32 then
		local newscale = (cursorx-framex)/HatMenu.ScalingWidth
		HatMenu.ScaleFrame(newscale)
	end
end

--[[ Cooldowns ]]

function HatMenu.UpdateWornCooldowns(maybeGlobal)
	local start,duration,enable = GetInventoryItemCooldown("player",1)
	CooldownFrame_SetTimer(HatMenu_HatCooldown,start,duration,enable)
	if not maybeGlobal then
		HatMenu.WriteWornCooldowns()
	end
end

function HatMenu.UpdateMenuCooldowns()
	local start,duration,enable
	for i=1,HatMenu.NumberOfHats do
		start,duration,enable = GetContainerItemCooldown(HatMenu.BaggedHats[i].bag,HatMenu.BaggedHats[i].slot)
		CooldownFrame_SetTimer(getglobal("HatMenu_Menu"..i.."Cooldown"),start,duration,enable)
	end
	HatMenu.WriteMenuCooldowns()
end

--[[ Item use ]]

function HatMenu.ReflectHatUse()
	getglobal("HatMenu_Hat"):SetChecked(1)
	HatMenu.StartTimer("UpdateWornHats")
	local _,_,id = string.find(GetInventoryItemLink("player",1) or "","item:(%d+)")
	if id then
		HatMenuPerOptions.ItemsUsed[id] = 0 -- 0 is an indeterminate state, cooldown will figure if it's worth watching
	end
end

function HatMenu.newUseInventoryItem(slot)
	if slot==1 and not MerchantFrame:IsVisible() then
		HatMenu.ReflectHatUse()
	end
end

function HatMenu.newUseAction(slot,cursor,self)
	if IsEquippedAction(slot) then
		HatMenu_TooltipScan:SetOwner(WorldFrame,"ANCHOR_NONE")
		HatMenu_TooltipScan:SetAction(slot)
		local _,hat = HatMenu.ItemInfo(1)
		if GameTooltipTextLeft1:GetText()==hat then
			HatMenu.ReflectHatUse()
		end
	end
end

--[[ Tooltips ]]

function HatMenu.WornHatTooltip()
	if HatMenu.IsTimerActive("Scaling") or HatMenuOptions.ShowTooltips=="OFF" then
		return
	end
	HatMenu.TooltipOwner = this
	HatMenu.TooltipType = "INVENTORY"
	HatMenu.TooltipBag = HatMenu.CombatQueue[0]
	HatMenu.StartTimer("TooltipUpdate",0)
end

function HatMenu.MenuHatTooltip()
	local id = this:GetID()
	if HatMenu.IsTimerActive("Scaling") or HatMenuOptions.ShowTooltips=="OFF" then
		return
	end
	HatMenu.TooltipOwner = this
	HatMenu.TooltipType = "BAG"
	HatMenu.TooltipBag = HatMenu.BaggedHats[id].bag
	HatMenu.TooltipSlot = HatMenu.BaggedHats[id].slot
	HatMenu.StartTimer("TooltipUpdate",0)
end

function HatMenu.ClearTooltip()
	GameTooltip:Hide()
	HatMenu.StopTimer("TooltipUpdate")
	HatMenu.TooltipType = nil
end

function HatMenu.AnchorTooltip(owner)
	owner = owner or this
	if HatMenuOptions.TooltipFollow=="ON" then
		if owner.GetLeft and owner:GetLeft() and owner:GetLeft()<400 then
			GameTooltip:SetOwner(owner,"ANCHOR_RIGHT")
		else
			GameTooltip:SetOwner(owner,"ANCHOR_LEFT")
		end
	else
		GameTooltip_SetDefaultAnchor(GameTooltip,owner)
	end
end

-- updates the tooltip created in the functions above
function HatMenu.TooltipUpdate()
	if HatMenu.TooltipType then
		local cooldown
		HatMenu.AnchorTooltip(HatMenu.TooltipOwner)
		if HatMenu.TooltipType=="BAG" then
			GameTooltip:SetBagItem(HatMenu.TooltipBag,HatMenu.TooltipSlot)
			cooldown = GetContainerItemCooldown(HatMenu.TooltipBag,HatMenu.TooltipSlot)
		else
			GameTooltip:SetInventoryItem("player",1)
			cooldown = GetInventoryItemCooldown("player",1)
		end
		HatMenu.ShrinkTooltip(HatMenu.TooltipOwner) -- if TinyTooltips on, shrink it
		if HatMenu.TooltipType=="INVENTORY" and HatMenu.TooltipBag then
			GameTooltip:AddLine("Queued: "..HatMenu.TooltipBag)
		end
		GameTooltip:Show()
		if cooldown==0 then
			-- stop updates if this hat has no cooldown
			HatMenu.TooltipType = nil
			HatMenu.StopTimer("TooltipUpdate")
		end
	end

end

-- normal tooltip for options
function HatMenu.OnTooltip(line1,line2)
	if HatMenuOptions.ShowTooltips=="ON" then
		HatMenu.AnchorTooltip()
		if line1 then
			GameTooltip:AddLine(line1)
			GameTooltip:AddLine(line2,.8,.8,.8,1)
			GameTooltip:Show()
		else
			local name = this:GetName() or ""
			for i=1,#(HatMenu.CheckOptInfo) do
				if name=="HatMenu_Opt"..HatMenu.CheckOptInfo[i][1] and HatMenu.CheckOptInfo[i][3] then
					HatMenu.OnTooltip(HatMenu.CheckOptInfo[i][3],HatMenu.CheckOptInfo[i][4])
				end
			end
			for i=1,#(HatMenu.TooltipInfo) do
				if HatMenu.TooltipInfo[i][1]==name and HatMenu.TooltipInfo[i][2] then
					HatMenu.OnTooltip(HatMenu.TooltipInfo[i][2],HatMenu.TooltipInfo[i][3])
				end
			end
		end
	end
end

-- strip format reordering in global strings
HatMenu.ITEM_SPELL_CHARGES = string.gsub(ITEM_SPELL_CHARGES,"%%%d%$d","%%d")
HatMenu.ITEM_SPELL_CHARGES_P1 = string.gsub(ITEM_SPELL_CHARGES_P1,"%%%d%$d","%%d")

function HatMenu.ShrinkTooltip(owner)
	if HatMenuOptions.TinyTooltips=="ON" then
		local r,g,b = GameTooltipTextLeft1:GetTextColor()
		local name = GameTooltipTextLeft1:GetText()
		local line,cooldown,charge
		for i=2,GameTooltip:NumLines() do
			line = getglobal("GameTooltipTextLeft"..i)
			if line:IsVisible() then
				line = line:GetText() or ""
				if string.find(line,COOLDOWN_REMAINING) then
					cooldown = line
				end
				if string.find(line,HatMenu.ITEM_SPELL_CHARGES_P1) or string.find(line,HatMenu.ITEM_SPELL_CHARGES) then
					charge = line
				end
			end
		end
		HatMenu.AnchorTooltip(owner or this)
		GameTooltip:AddLine(name,r,g,b)
		GameTooltip:AddLine(charge,1,1,1)
		GameTooltip:AddLine(cooldown,1,1,1)
	end
end

-- returns 1 if the item at bag(,slot) is an engineered hat
function HatMenu.IsEngineered(bag,slot)
	local item = slot and GetContainerItemLink(bag,slot) or GetInventoryItemLink("player",1)
	if item then
		local _,_,_,_,_,itemType,itemSubtype,_,itemLoc = GetItemInfo(item)
		if itemType==HatMenu.TRADE_GOODS and itemSubtype==HatMenu.DEVICES and itemLoc=="INVTYPE_HEAD" then
			return 1
		end
		HatMenu_TooltipScan:SetOwner(WorldFrame,"ANCHOR_NONE")
		HatMenu_TooltipScan:SetHyperlink(item)
		for i=1,HatMenu_TooltipScan:NumLines() do
			if string.match(getglobal("HatMenu_TooltipScanTextLeft"..i):GetText() or "",HatMenu.REQUIRES_ENGINEERING) then
				return 1
			end
		end
	end
end

-- returns bag,slot of a free bag space, if one found.  engineering true if only looking for an engineering bag
function HatMenu.FindSpace(engineering)
	local bagType
	for i=4,0,-1 do
		bagType = select(7,GetItemInfo(GetInventoryItemLink("player",19+i) or ""))
		if (engineering and bagType==HatMenu.ENGINEERING_BAG) or (not engineering and bagType==HatMenu.BAG) then
			for j=1,GetContainerNumSlots(i) do
				if not GetContainerItemLink(i,j) then
					return i,j
				end
			end
		end
	end
end

--[[ Combat Queue ]]

function HatMenu.EquipHatByName(name)
	if not name then return end
	if UnitAffectingCombat("player") or HatMenu.IsPlayerReallyDead() then
		-- queue hat
		local queue = HatMenu.CombatQueue
		if queue[0]==name and not imperative then
			queue[0] = nil
		else
			queue[0] = name
		end
	elseif not CursorHasItem() and not SpellIsTargeting() then
		local _,b,s = HatMenu.FindItem(name)
		if b then
			local _,_,isLocked = GetContainerItemInfo(b,s)
			if not isLocked and not IsInventoryItemLocked(1) then
				-- neither container item nor inventory item locked, perform swap
				local directSwap = 1 -- assume a direct swap will happen
				if select(7,GetItemInfo(GetInventoryItemLink("player",19+b) or ""))==HatMenu.ENGINEERING_BAG then
					-- incoming hat is in an engineering bag
					if not HatMenu.IsEngineered(1) then
						-- outgoing hat can't go inside it
						local freeBag,freeSlot = HatMenu.FindSpace()
						if freeBag then
							PickupInventoryItem(1)
							PickupContainerItem(freeBag,freeSlot)
							PickupContainerItem(b,s)
							EquipCursorItem(1)
							directSwap = nil
						end
					end
				elseif HatMenu.IsEngineered(1) and not HatMenu.IsEngineered(b,s) then
					-- outgoing hat is engineered, incoming hat is not
					local freeBag,freeSlot = HatMenu.FindSpace(1)
					if freeBag then
						-- move outgoing hat to engineering bag, equip incoming hat
						PickupInventoryItem(1)
						PickupContainerItem(freeBag,freeSlot)
						PickupContainerItem(b,s)
						EquipCursorItem(1)
						directSwap = nil
					end
				end
				if directSwap then
					PickupContainerItem(b,s)
					PickupInventoryItem(1)
				end
				getglobal("HatMenu_HatIcon"):SetDesaturated(1)
				HatMenu.StartTimer("UpdateWornHats") -- in case it's not equipped (stunned, etc)
			end
		end
	end
	HatMenu.UpdateCombatQueue()
end	

function HatMenu.UpdateCombatQueue()
	local bag,slot
	local hat = HatMenu.CombatQueue[0]
	local icon = getglobal("HatMenu_HatQueue")
	icon:Hide()
	if hat then
		_,bag,slot = HatMenu.FindItem(hat)
		if bag then
			icon:SetTexture(GetContainerItemInfo(bag,slot))
			icon:Show()
		end
	end
end

--[[ Notify ]]

function HatMenu.Notify(msg)
	PlaySound("GnomeExploration")
	if SCT_Display then -- send via SCT if it exists
		SCT_Display(msg,{r=.2,g=.7,b=.9})
	elseif SHOW_COMBAT_TEXT=="1" then
		CombatText_AddMessage(msg, CombatText_StandardScroll, .2, .7, .9) -- or default UI's SCT
	else
		-- send vis UIErrorsFrame if neither SCT exists
		UIErrorsFrame:AddMessage(msg,.2,.7,.9,1,UIERRORS_HOLD_TIME)
	end
	if HatMenuOptions.NotifyChatAlso=="ON" then
		DEFAULT_CHAT_FRAME:AddMessage("|cff33b2e5"..msg)
	end
end

function HatMenu.CooldownUpdate()
	local inv,bag,slot,start,duration,name,remain
	for i in pairs(HatMenuPerOptions.ItemsUsed) do
		start,duration = GetItemCooldown(i)
		if start and HatMenuPerOptions.ItemsUsed[i]<3 then
			HatMenuPerOptions.ItemsUsed[i] = HatMenuPerOptions.ItemsUsed[i] + 1 -- count for 3 seconds before seeing if this is a real cooldown
		elseif start then
			if start>0 then
				remain = duration - (GetTime()-start)
				if HatMenuPerOptions.ItemsUsed[i]<5 then
					if remain>29 then
						HatMenuPerOptions.ItemsUsed[i] = 30 -- first actual cooldown greater than 30 seconds, tag it for 30+0 notify
					elseif remain>5 then
						HatMenuPerOptions.ItemsUsed[i] = 5 -- first actual cooldown less than 30 but greater than 5, tag for 0 notify
					end
				end
			end
			if HatMenuPerOptions.ItemsUsed[i]==30 and start>0 and remain<30 then
				if HatMenuOptions.NotifyThirty=="ON" then
					name = GetItemInfo(i)
					if name then
						HatMenu.Notify(name.." ready soon!")
					end
				end
				HatMenuPerOptions.ItemsUsed[i]=5 -- tag for just 0 notify now
			elseif HatMenuPerOptions.ItemsUsed[i]==5 and start==0 then
				if HatMenuOptions.Notify=="ON" then
					name = GetItemInfo(i)
					if name then
						HatMenu.Notify(name.." ready!")
					end
				end
			end
			if start==0 then
				HatMenuPerOptions.ItemsUsed[i] = nil
			end
		end
	end

	-- update cooldown numbers
	if HatMenuOptions.CooldownCount=="ON" then
		if HatMenu_MainFrame:IsVisible() then
			HatMenu.WriteWornCooldowns()
		end
		if HatMenu_MenuFrame:IsVisible() then
			HatMenu.WriteMenuCooldowns()
		end
	end
end

function HatMenu.WriteWornCooldowns()
	local start, duration
	start, duration = GetInventoryItemCooldown("player",1)
	HatMenu.WriteCooldown(HatMenu_HatTime,start,duration)
end

function HatMenu.WriteMenuCooldowns()
	local start, duration
	for i=1,HatMenu.NumberOfHats do
		start, duration = GetContainerItemCooldown(HatMenu.BaggedHats[i].bag,HatMenu.BaggedHats[i].slot)
		HatMenu.WriteCooldown(getglobal("HatMenu_Menu"..i.."Time"),start,duration)
	end
end

function HatMenu.WriteCooldown(where,start,duration)
	local cooldown = duration - (GetTime()-start)
	if start==0 or HatMenuOptions.CooldownCount=="OFF" then
		where:SetText("")
	elseif cooldown<3 and not where:GetText() then
		-- this is a global cooldown. don't display it. not accurate but at least not annoying
	else
		where:SetText((cooldown<60 and math.floor(cooldown+.5).." s") or (cooldown<3600 and math.ceil(cooldown/60).." m") or math.ceil(cooldown/3600).." h")
	end
end

function HatMenu.OnShow()
	HatMenuPerOptions.Visible = "ON"
	if HatMenuOptions.KeepOpen=="ON" then
		HatMenu.BuildMenu()
	end
end

function HatMenu.OnHide()
	HatMenuPerOptions.Visible = "OFF"
	HatMenu_MenuFrame:Hide()
end

function HatMenu.ReflectAlpha()
	HatMenu_MainFrame:SetAlpha(HatMenuPerOptions.Alpha)
	HatMenu_MenuFrame:SetAlpha(HatMenuPerOptions.Alpha)
end

--[[ Key bindings ]]

function HatMenu.KeyBindingsChanged()
	if HatMenuOptions.ShowHotKeys=="ON" then
		local key = GetBindingKey("CLICK HatMenu_Hat:LeftButton")
		getglobal("HatMenu_HatHotKey"):SetText(GetBindingText(key or "",nil,1))
	else
		HatMenu_HatHotKey:SetText("")
	end
end

--[[ Monitor Range ]]

function HatMenu.ReflectRedRange()
	if HatMenuOptions.RedRange=="ON" then
		HatMenu.StartTimer("RedRange")
	else
		HatMenu.StopTimer("RedRange")
		HatMenu_HatIcon:SetVertexColor(1,1,1)
	end
end

function HatMenu.RedRangeUpdate()
	local item = GetInventoryItemLink("player",1)
	if item and IsItemInRange(item)==0 then
		getglobal("HatMenu_HatIcon"):SetVertexColor(1,.3,.3)
	else
		getglobal("HatMenu_HatIcon"):SetVertexColor(1,1,1)
	end
end
