--[[ HatMenuOpt.lua : Options and sort window for HatMenu ]]

HatMenu.CheckOptInfo = {
{"ShowIcon","ON","Minimap button","Show or hide minimap button."},
{"SquareMinimap","OFF","Square minimap","Move minimap button as if around a square minimap.","ShowIcon"},
{"CooldownCount","OFF","Cooldown numbers","Display time remaining on cooldowns ontop of the button."},
{"TooltipFollow","OFF","At cursor","Display all tooltips at the cursor.","ShowTooltips"},
{"KeepOpen","OFF","Keep menu open","Keep menu open at all times."},
{"KeepDocked","ON","Keep menu docked","Keep menu docked at all times."},
{"Notify","OFF","Notify when ready","Sends an overhead notification when a hat's cooldown is complete."},
{"DisableToggle","OFF","Disable toggle","Disables the minimap button's ability to toggle the hat frame.","ShowIcon"},
{"NotifyChatAlso","OFF","Notify chat also","Sends notifications through chat also."},
{"Locked","OFF","Lock windows","Prevents the windows from being moved, resized or rotated."},
{"ShowTooltips","ON","Show tooltips","Shows tooltips."},
{"NotifyThirty","ON","Notify at 30 sec","Sends an overhead notification when a hat has 30 seconds left on cooldown."},
{"MenuOnShift","OFF","Menu on shift","Check this to prevent the menu appearing unless Shift is held."},
{"TinyTooltips","OFF","Tiny tooltips","Shrink hat tooltips to only their name, charges and cooldown.","ShowTooltips"},
{"SetColumns","OFF","Set menu columns","Define how many hats before the menu will wrap to the next row.\n\nUncheck to let HatMenu choose how to wrap the menu."},
{"LargeCooldown","ON","Large numbers","Display the cooldown time in a larger font.","CooldownCount"},
{"ShowHotKeys","ON","Show key bindings","Display the key bindings over the equipped hats."},
{"RedRange","OFF","Red out of range","Check this to red out worn hats that are out of range to a valid target.  ie, Goblin Rocket Helmet."},
{"MenuOnRight","OFF","Menu on right-click","Check this to prevent the menu from appearing until either worn hat is right-clicked.\n\nNOTE: This setting CANNOT be changed while in combat."}
}

-- table.insert(HatMenu.CheckOptInfo,)

HatMenu.TooltipInfo = {
{"HatMenu_LockButton","Lock windows","Prevents the windows from being moved, resized or rotated."},
{"HatMenu_OptBindButton","Bind hat","Click this to bind a key to the hat."}
}

function HatMenu.InitOptions()
	HatMenu.CreateTimer("DragMinimapButton",HatMenu.DragMinimapButton,0,1)
	HatMenu.MoveMinimapButton()
	local item
	for i=1,#(HatMenu.CheckOptInfo) do
		item = getglobal("HatMenu_Opt"..HatMenu.CheckOptInfo[i][1].."Text")
		if item then
			item:SetText(HatMenu.CheckOptInfo[i][3])
			item:SetTextColor(.95,.95,.95)
		end
	end
	table.insert(UISpecialFrames,"HatMenu_OptFrame")
	HatMenu_Title:SetText("HatMenu "..HatMenu_Version)

	HatMenu_OptFrame:SetBackdropBorderColor(.3,.3,.3,1)
	HatMenu_OptFrame:SetHeight(306)
	HatMenu_SubOptFrame:SetBackdropBorderColor(.3,.3,.3,1)
	HatMenu_SubOptFrame:SetPoint("TOPLEFT",HatMenu_OptFrame,"TOPLEFT",8,-24)
	HatMenu_OptColumnsSlider:SetValue(HatMenuOptions.Columns)
	HatMenu.ReflectLock()
	HatMenu.ReflectCooldownFont()
	HatMenu.KeyBindingsChanged()
end

function HatMenu.ToggleFrame(frame)
	if frame:IsVisible() then
		frame:Hide()
	else
		frame:Show()
	end
end

function HatMenu.OptFrame_OnShow()
	HatMenu.ValidateChecks()
end

--[[ Minimap button ]]

function HatMenu.MoveMinimapButton()
	local xpos,ypos
	if HatMenuOptions.SquareMinimap=="ON" then
		xpos = 110 * cos(HatMenuOptions.IconPos or 0)
		ypos = 110 * sin(HatMenuOptions.IconPos or 0)
		xpos = math.max(-82,math.min(xpos,84))
		ypos = math.max(-86,math.min(ypos,82))
	else
		xpos = 80*cos(HatMenuOptions.IconPos or 0)
		ypos = 80*sin(HatMenuOptions.IconPos or 0)
	end
	HatMenu_IconFrame:SetPoint("TOPLEFT","Minimap","TOPLEFT",52-xpos,ypos-52)
	if HatMenuOptions.ShowIcon=="ON" then
		HatMenu_IconFrame:Show()
	else
		HatMenu_IconFrame:Hide()
	end
end

function HatMenu.DragMinimapButton()
	local xpos,ypos = GetCursorPosition()
	local xmin,ymin = Minimap:GetLeft() or 400, Minimap:GetBottom() or 400
	xpos = xmin-xpos/Minimap:GetEffectiveScale()+70
	ypos = ypos/Minimap:GetEffectiveScale()-ymin-70
	HatMenuOptions.IconPos = math.deg(math.atan2(ypos,xpos))
	HatMenu.MoveMinimapButton()
end

function HatMenu.MinimapButton_OnClick()
	PlaySound("GAMEGENERICBUTTONPRESS")
	if IsShiftKeyDown() then
		HatMenuOptions.Locked = HatMenuOptions.Locked=="ON" and "OFF" or "ON"
		HatMenu.ReflectLock()
	else
		if arg1=="LeftButton" and HatMenuOptions.DisableToggle=="OFF" then
			HatMenu.ToggleFrame(HatMenu_MainFrame)
		else
			HatMenu.ToggleFrame(HatMenu_OptFrame)
		end
	end
end

--[[ CheckButton ]]

function HatMenu.ValidateChecks()
	local check,button
	for i=1,#(HatMenu.CheckOptInfo) do
		check = HatMenu.CheckOptInfo[i]
		button = getglobal("HatMenu_Opt"..check[1])
		if button then
			button:SetChecked(HatMenuOptions[check[1]]=="ON")
			if check[5] then
				if HatMenuOptions[check[5]]=="ON" then
					button:Enable()
					getglobal("HatMenu_Opt"..check[1].."Text"):SetTextColor(.95,.95,.95)
				else
					button:Disable()
					getglobal("HatMenu_Opt"..check[1].."Text"):SetTextColor(.5,.5,.5)
				end
			end
		end
	end
	HatMenu_OptColumnsSlider:SetAlpha((HatMenuOptions.SetColumns=="ON") and 1 or .5)
	HatMenu_OptColumnsSlider:EnableMouse((HatMenuOptions.SetColumns=="ON") and 1 or 0)
	HatMenu_OptColumnsSlider:SetValue(HatMenuOptions.Columns)
end

function HatMenu.OptColumnsSlider_OnValueChanged()
	if HatMenuOptions then
		HatMenuOptions.Columns = this:GetValue()
		HatMenu_OptColumnsSliderText:SetText(HatMenuOptions.Columns.." hats")
		if HatMenu_MenuFrame:IsVisible() then
			HatMenu.BuildMenu()
		end
	end
end

function HatMenu.CheckButton_OnClick()
	local _,_,var = string.find(this:GetName(),"HatMenu_Opt(.+)")
	if HatMenuOptions[var] then
		HatMenuOptions[var] = this:GetChecked() and "ON" or "OFF"
		PlaySound(this:GetChecked() and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
		HatMenu.ValidateChecks()
	end

	if this==HatMenu_OptCooldownCount then
		HatMenu.WriteWornCooldowns()
		HatMenu.WriteMenuCooldowns()
	elseif this==HatMenu_OptLocked then
		HatMenu.DockWindows()
		HatMenu.ReflectLock()
	elseif this==HatMenu_OptKeepOpen or this==HatMenu_OptSetColumns then
		if HatMenuOptions.KeepOpen=="ON" then
			HatMenu.BuildMenu()
		end
	elseif this==HatMenu_OptKeepDocked then
		HatMenu.DockWindows()
	elseif this==HatMenu_OptLargeCooldown then
		HatMenu.ReflectCooldownFont()
	elseif this==HatMenu_OptSquareMinimap then
		HatMenu.MoveMinimapButton()
	elseif this==HatMenu_OptShowHotKeys then
		HatMenu.KeyBindingsChanged()
	elseif this==HatMenu_OptShowIcon then
		HatMenu.MoveMinimapButton()
	elseif this==HatMenu_OptRedRange then
		HatMenu.ReflectRedRange()
	elseif this==HatMenu_OptMenuOnRight then
		HatMenu.ReflectMenuOnRight()
	end
end

function HatMenu.ReflectLock()
	local c = HatMenuOptions.Locked=="ON" and 0 or .5
	HatMenu_OptFrame:SetBackdropBorderColor(c,c,c,1)
	HatMenu_MainFrame:SetBackdropColor(c,c,c,c)
	HatMenu_MainFrame:SetBackdropBorderColor(c,c,c,c*2)
	HatMenu_MenuFrame:SetBackdropColor(c,c,c,c)
	HatMenu_MenuFrame:SetBackdropBorderColor(c,c,c,c*2)
	HatMenu_MenuFrame:EnableMouse(c*2)
	HatMenu_OptLocked:SetChecked(1-c*2)
	local normalTexture = HatMenu_LockButton:GetNormalTexture()
	local pushedTexture = HatMenu_LockButton:GetPushedTexture()
	if c==0 then
		HatMenu_MainResizeButton:Hide()
		HatMenu_MenuResizeButton:Hide()
		normalTexture:SetTexCoord(.875,1,.125,.25)
		pushedTexture:SetTexCoord(.75,.875,.125,.25)
	else
		HatMenu_MainResizeButton:Show()
		HatMenu_MenuResizeButton:Show()
		normalTexture:SetTexCoord(.75,.875,.125,.25)
		pushedTexture:SetTexCoord(.875,1,.125,.25)
	end
end

function HatMenu.ReflectCooldownFont()
	HatMenu.SetCooldownFont("HatMenu_Hat")
	for i=1,30 do
		HatMenu.SetCooldownFont("HatMenu_Menu"..i)
	end
end

function HatMenu.SetCooldownFont(button)
	local item = getglobal(button.."Time")
	if HatMenuOptions.LargeCooldown=="ON" then
		item:SetFont("Fonts\\FRIZQT__.TTF",16,"OUTLINE")
		item:SetTextColor(1,.82,0,1)
		item:ClearAllPoints()
		item:SetPoint("CENTER",button,"CENTER")
	else
		item:SetFont("Fonts\\ARIALN.TTF",14,"OUTLINE")
		item:SetTextColor(1,1,1,1)
		item:ClearAllPoints()
		item:SetPoint("BOTTOM",button,"BOTTOM")
	end
end


--[[ Titlebar buttons ]]

function HatMenu.SmallButton_OnClick()
	PlaySound("igMainMenuOptionCheckBoxOn")
	if this==HatMenu_CloseButton then
		HatMenu_OptFrame:Hide()
	elseif this==HatMenu_LockButton then
		HatMenuOptions.Locked = (HatMenuOptions.Locked=="ON") and "OFF" or "ON"
		HatMenu.DockWindows()
		HatMenu.ReflectLock()
	end
end


function HatMenu.BindButton_OnClick()
	local bindings = {
		{"Use hat","CLICK HatMenu_Hat:LeftButton"},
	}
	ClickBinder.Toggle("HatMenu",bindings,HatMenu_OptFrame)
end
